/*********************************************************************************************************/
/*** 			                    	Benjamin Fletcher - bjf1g13                                    ***/
/*** 	                        Power Neutral Energy Harvesting Management                             ***/
/*** 	                              for Multicore Embedded systems                                   ***/
/*** 	                                                                                               ***/
/*** 	                                       core_switch.h                                           ***/
/*** 	                                                                                               ***/
/*** 	                              (see report for documentation)                                   ***/
/*********************************************************************************************************/


#define SYSFS_GPIO_DIR  "/sys/class/gpio"
#define VALUE_LENGTH         8
#define INSTRUCT_LENGTH		64
#define MAX_BUF 64

void export_gpio(int gpio_number)
{
	int fd;
	char value[VALUE_LENGTH] = {};
	fd = open("/sys/class/gpio/export", O_WRONLY);
	sprintf(value, "%d", gpio_number);	
	write(fd, &value, VALUE_LENGTH);
	close(fd);
}

void unexport_gpio(int gpio_number)
{
	int fd;
	char value[VALUE_LENGTH] = {};
	fd = open("/sys/class/gpio/unexport", O_WRONLY);
	sprintf(value, "%d", gpio_number);	
	write(fd, &value, VALUE_LENGTH);
	close(fd);
}

void gpio_direction(int gpio_number, bool output)
{
	int fd;
	char instruction[INSTRUCT_LENGTH] = {};
	sprintf(instruction, "/sys/class/gpio/gpio%d/direction", gpio_number);
	fd = open(instruction, O_WRONLY);
		
	if (output)
		write(fd, "out", 4);
	else
		write(fd, "in", 3);
	close(fd);
}

void gpio_get(int gpio_number, int * value)
{
	int fd;
	char instruction[INSTRUCT_LENGTH] = {};
	char return_value;
	sprintf(instruction, "/sys/class/gpio/gpio%d/value", gpio_number);
	fd = open(instruction, O_RDONLY);
	
	read(fd, &return_value, 1);
		
	if(return_value == '0')
		*value = 0;
	else
		*value = 1;
	close(fd);
}

void gpio_set(int gpio_number, bool value)
{
	int fd;
	char instruction[INSTRUCT_LENGTH] = {};
	sprintf(instruction, "/sys/class/gpio/gpio%d/value", gpio_number);
	fd = open(instruction, O_WRONLY);
		
	if (value)
		write(fd, "1", 2);
	else
		write(fd, "0", 2);
	close(fd);
}

void gpio_edge(int gpio_number, char * edge)
{
	int fd;
	char instruction[INSTRUCT_LENGTH] = {};
	sprintf(instruction, "/sys/class/gpio/gpio%d/edge", gpio_number);
	fd = open(instruction, O_WRONLY);
	
	write(fd, edge, strlen(edge) + 1);

	close(fd);
}

int open_gpio(int gpio_number)
{
	int fd;
	char instruction[INSTRUCT_LENGTH] = {};
	sprintf(instruction, "/sys/class/gpio/gpio%d/value", gpio_number);
	return open(instruction, O_RDONLY | O_NONBLOCK );
}

void close_gpio(int fd)
{
	close(fd);
}
 
void wait(void)
{	
	usleep(30000);
}


