/*****************************************************************/
/*** 				Benjamin Fletcher - bjf1g13                ***/
/*** 	    Power Neutral Energy Harvesting Management         ***/
/*** 	          for Multicore Embedded systems               ***/
/*** 	                                                       ***/
/*** 	                     main.cpp                          ***/
/*** 	                                                       ***/
/*** 	          (see report for documentation)               ***/
/*****************************************************************/

/* ------------------ Precompiled Headers --------------------- */
#include <iostream>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <ctime>  

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sched.h>
#include <poll.h>
#include <time.h>
#include <stdint.h>                                                             
#include <unistd.h>                                                            
#include <fcntl.h>                                                              
#include <sys/ioctl.h>                                                          
#include <sys/time.h>
#include <linux/spi/spidev.h>

/* ------------------ Downloaded Headers --------------------- */

#include "cpufreq.h"

/* ------------------------ My Headers ----------------------- */

#include "core_switch.h"
#include "frequency_switch.h"
#include "gpio.h"
#include "print.h"

/* ----------------- Algrithm Definitions -------------------- */

#define LOW_VOLTAGE_INTERRUPT_PIN     28 // Physical Number 20  
#define HIGH_VOLTAGE_INTERRUPT_PIN    19 // Physical Number 18
#define LOW_VOLTAGE_INTERRUPT_CS_PIN  24 // Physical Number 26
#define HIGH_VOLTAGE_INTERRUPT_CS_PIN 25 // Physical Number 24

#define LITTLE_CORE_SCALE_DELAY     2
#define BIG_CORE_SCALE_DELAY 1

#define INTERRUPT_WIDTH  4 
#define THRESHOLD_OFFSET 2

#define NO_OF_THRESHOLDS 2
#define POT_RESOLUTION  128

#define PROGRAM_TIMEOUT 5000

#define POLL_TIMEOUT    100 // 100ms
#define BUFFER_LEN      64

/* ----------------- Set Operating Frequencies ---------------- */

static int op_freq[] = { 200000000,  // 0
			             450000000,  // 1
			             720000000,  // 2
			             920000000,  // 3
			             1100000000, // 4
			             1200000000, // 5
			             1300000000, // 6
			             1400000000};// 7

int current_freq                = 4;
int current_active_little_cores = 3;
int current_active_big_cores    = 0;
int current_high_interrupt      = 0;
int current_low_interrupt       = 0;


/* --------------------------- SPI Stuff ---------------------------- */  

char           spi_device[] = "/dev/spidev1.0";                                   
unsigned int   spi_bits     = 8;                                                        
unsigned int   spi_speed   = 500;                                                 
 
int spi_trans(int spi_file, char addr, char data) {

    uint8_t tx[] = {addr, data};
    uint8_t rx[sizeof(tx) / sizeof(tx[0])] = {};
 
    struct spi_ioc_transfer tr;
 
    tr.tx_buf        = (unsigned long)tx;
    tr.rx_buf        = (unsigned long)rx;
    tr.speed_hz      = spi_speed;
    tr.bits_per_word = spi_bits;
    tr.len           = sizeof(tx) / sizeof(tx[0]);
 
    return ioctl(spi_file, SPI_IOC_MESSAGE(1), &tr);

} 

//* ---------------------------  Initialisation ---------------------------- */   

int initialise_test(void)
{
	set_A_cores_frequency(op_freq[current_freq]);
	set_B_cores_frequency(op_freq[current_freq]);
	
	for(int n=0; n<8; n++)
		turn_off_core(n);
	
	// Init with 3 cores online 
	turn_on_core(1);
	turn_on_core(2);

	return 0;
}

/* ------------------------  Function Definitions -------------------------- */  

void decreace_frequency(void) 
{
	if(current_freq > 0)
	{
		set_B_cores_frequency(op_freq[current_freq - 1]);
		current_freq --;
	}
}

void increace_frequency(void)
{
	if(current_freq < 7)
	{
		set_B_cores_frequency(op_freq[current_freq + 1]);
		current_freq ++;
	}
}

void add_core(bool big_core)
{
	if(big_core == 1)
	{
		if(current_active_big_cores < 4)
		{
			turn_on_core(current_active_big_cores +4);
			current_active_big_cores ++;
			print_hotplugging_message(1,1);
		}
		else if(current_active_little_cores < 4)
		{
			turn_on_core(current_active_little_cores);
			current_active_little_cores ++;
			print_hotplugging_message(1,0);
		}
		else
			print_max_limit(1);
	}
	else
	{	
		if(current_active_little_cores < 4)
		{
			turn_on_core(current_active_little_cores);
			current_active_little_cores ++;
			print_hotplugging_message(1,0);
		}
		else
			print_max_limit(0);
	}
}

void remove_core(bool big_core)
{
	if(big_core == 1)
	{
		if(current_active_big_cores >1)
		{
			current_active_big_cores --;
			turn_off_core(current_active_big_cores +4);
			print_hotplugging_message(0,1);
		}
		else if(current_active_little_cores > 1)
		{
			current_active_little_cores --;
			turn_off_core(current_active_little_cores);
			print_hotplugging_message(0,0);
		}
		else
			print_min_limit(1);
	}
	else
	{	
		if(current_active_little_cores > 1)
		{
			current_active_little_cores --;
			turn_off_core(current_active_little_cores);
			print_hotplugging_message(0,0);
		}
		else
			print_min_limit(0);
	}

}

void set_threshold(int threshold, int gpio_cs)
{
	gpio_set(gpio_cs, 0);
	int fd_spi = open(spi_device, O_RDWR);  
	spi_trans(fd_spi, 0, threshold);
	close(fd_spi);
	gpio_set(gpio_cs, 1);
}

void increment_thresholds()
{
	current_low_interrupt  = current_low_interrupt  - THRESHOLD_OFFSET/2;
	current_high_interrupt = current_high_interrupt - THRESHOLD_OFFSET/2;
		
	set_threshold(current_low_interrupt,  LOW_VOLTAGE_INTERRUPT_CS_PIN );
	set_threshold(current_high_interrupt, HIGH_VOLTAGE_INTERRUPT_CS_PIN);
}

void decrement_thresholds()
{
	current_low_interrupt  = current_low_interrupt  + THRESHOLD_OFFSET/2;
	current_high_interrupt = current_high_interrupt + THRESHOLD_OFFSET/2;
		
	set_threshold(current_low_interrupt,  LOW_VOLTAGE_INTERRUPT_CS_PIN );
	set_threshold(current_high_interrupt, HIGH_VOLTAGE_INTERRUPT_CS_PIN);
}



/* -----------------------------------  Main Algorithm --------------------------- */  

int main(void)
{
	char   trigger_high[] = "rising";
	char   trigger_low[]  = "falling";
	
	int    number_of_thresholds = NO_OF_THRESHOLDS;
	int    gpio_fd_low, gpio_fd_high, gpio_h_cs, gpio_l_cs, timeout, poll_status, len;
	char   *buf[BUFFER_LEN];
	unsigned int gpio_low, gpio_high;
	bool currently_decreacing = 0;
	bool currently_increacing = 0;
	unsigned int * pin_value_low;
	unsigned int * pin_value_high;
	int global_timer      = 0;
	int high_timer        = 0;
	int low_timer         = 0;
	int new_mid_threshold = 0;
	int return_value      = 0;
	struct timeval tv;
	struct pollfd poll_file_s[NO_OF_THRESHOLDS];

	gpio_low  = LOW_VOLTAGE_INTERRUPT_PIN;
	gpio_high = HIGH_VOLTAGE_INTERRUPT_PIN;
	gpio_h_cs = HIGH_VOLTAGE_INTERRUPT_CS_PIN;
	gpio_l_cs = LOW_VOLTAGE_INTERRUPT_CS_PIN;
	
	close_gpio(gpio_low);
	close_gpio(gpio_high);
	unexport_gpio(gpio_low);
	unexport_gpio(gpio_high);
	unexport_gpio(gpio_h_cs);
	unexport_gpio(gpio_l_cs);  
                           
	export_gpio(gpio_low);                // Export the low Interrupt GPIO Pin
	gpio_direction(gpio_low, 0);          // Set at Input
	gpio_edge(gpio_low, trigger_low);     // Set both edges trigger
	gpio_fd_low = open_gpio(gpio_low);    // Open

	export_gpio(gpio_high);
	gpio_direction(gpio_high, 0);
	gpio_edge(gpio_high, trigger_high);
	gpio_fd_high = open_gpio(gpio_high);

	export_gpio(gpio_h_cs);
	gpio_direction(gpio_h_cs, 1);

	export_gpio(gpio_l_cs);
	gpio_direction(gpio_l_cs, 1);
	
	timeout = POLL_TIMEOUT;

	print_headers();
	print_pre_core_switch_message();
	frequency_switch_init();
	initialise_test();
	print_post_core_switch_message();

	int init_threshold_low = 0;
	int init_threshold_high = 0;
	int value_low;
	int value_high;

	gpio_set(gpio_h_cs, 1);
	gpio_set(gpio_l_cs, 1); 

	int high_threshold = 0;
	int low_threshold  = 0;

	std::ofstream log_file;
	log_file.open("log.csv");

	// Low Threshold Initialisation
	for(int n = 0; n < POT_RESOLUTION; n ++)
	{
		if(init_threshold_low == 0)
		{
			set_threshold(n, gpio_l_cs);
			wait();
			gpio_get(gpio_low, &value_low);
			if(value_low == 1)
			{
				init_threshold_low = n;
			}
		}
	}

	// Set Initial High Threshold Value
	current_low_interrupt = init_threshold_low + INTERRUPT_WIDTH/2;
	set_threshold(current_low_interrupt, gpio_l_cs);
	wait();

	// High Threshold Initialisation
	for(int n = 0; n < POT_RESOLUTION; n ++)
	{
		if(init_threshold_high == 0)
		{
			set_threshold(n, gpio_h_cs);
			wait();
			gpio_get(gpio_high, &value_high);
			if(value_high == 1)
			{
				init_threshold_high = n;
			}
		}
	}

	// Set Initial High Threshold Value
	current_high_interrupt= init_threshold_high - INTERRUPT_WIDTH/2;
	set_threshold(current_low_interrupt, gpio_h_cs);

	print_success(init_threshold_high, init_threshold_low);
	wait();


while (global_timer < PROGRAM_TIMEOUT) 
{
		memset((void*)poll_file_s, 0, sizeof(poll_file_s));

		poll_file_s[1].fd = gpio_fd_low;
		poll_file_s[1].events = POLLPRI;

		poll_file_s[0].fd = gpio_fd_high;
		poll_file_s[0].events = POLLPRI;
		
		poll_status = poll(poll_file_s, number_of_thresholds, timeout);      
     
		if (poll_status == 0) // 0.1s Timeout has occured
		{
			global_timer ++;
			high_timer   ++;
			low_timer    ++;
		}

		if (poll_file_s[0].revents & POLLPRI)            // High INT has occured
		{ 
			global_timer ++;                             // Increment Gloabl Event Counter
			read(poll_file_s[0].fd, buf, BUFFER_LEN); 	 // Read the updated data from the file to reset
			print_interrupt_message(global_timer, 1);    // Print Message that High interrupt has occured

			increment_thresholds();
			increace_frequency();
			print_frequency_message(global_timer, 1, op_freq[current_freq]); // Print Message that Frequency has been altered

			if(high_timer < BIG_CORE_SCALE_DELAY)                            // Calculate core scaling response
				add_core(1);
			else if(high_timer < LITTLE_CORE_SCALE_DELAY)
				add_core(0);

			high_timer = 0; 										 		 // Reset Timer

			// Create a timestamped log of the event
			print_to_log(log_file, 1, current_active_little_cores, current_active_big_cores, op_freq[current_freq]);			

		}
		
		
		// Low Voltage Interrupt on either side has occured
		if (poll_file_s[1].revents & POLLPRI) 
		{
			global_timer ++;                             // Increment Gloabl Event Counter
			read(poll_file_s[1].fd, buf, BUFFER_LEN); 		 // Read the updated data from the file to reset
			print_interrupt_message(global_timer, 0);    // Print Message that High interrupt has occured

			decrement_thresholds();
			decreace_frequency();

			print_frequency_message(global_timer, 0, op_freq[current_freq]); // Print Message that Frequency has been altered
				
			if(low_timer < BIG_CORE_SCALE_DELAY) 							 // Calculate core scaling response
				remove_core(1);
			else if(low_timer < LITTLE_CORE_SCALE_DELAY)
				remove_core(0);

			low_timer = 0; 										 		     // Reset Timer
	
			// Create a timestamped log of the event
			print_to_log(log_file, 0, current_active_little_cores, current_active_big_cores, op_freq[current_freq]);

		}
	}

	print_exit_message();
	
	// Close Opened Files //

	close_gpio(gpio_low);
	close_gpio(gpio_high);
	unexport_gpio(gpio_low);
	unexport_gpio(gpio_high);
	unexport_gpio(gpio_h_cs);
	unexport_gpio(gpio_l_cs); 
	log_file.close();                            
                                                          
    return 0; 
}                                                                   
 
                                                                                                              
