/*********************************************************************************************************/
/*** 			                    	Benjamin Fletcher - bjf1g13                                    ***/
/*** 	                        Power Neutral Energy Harvesting Management                             ***/
/*** 	                              for Multicore Embedded systems                                   ***/
/*** 	                                                                                               ***/
/*** 	                                         print.h                                               ***/
/*** 	                                                                                               ***/
/*** 	                              (see report for documentation)                                   ***/
/*********************************************************************************************************/

/* ----------------------------------------   Parameters  -----------------------------------------------*/

#define VERBOSITY 2  

/* ----------------------------------------   Functions  ------------------------------------------------*/

void print_headers()
{
	std::cout << "#######################################################################" << std::endl
		 	  << "#   Power Neutral Energy Harvesting Management Algorithm for ODROID   #" << std::endl
		  	  << "#                                                                     #" << std::endl
		      << "#                 Ben Fletcher (bjf1g13@ecs.soton.ac.uk)              #" << std::endl
		  	  << "#######################################################################" << std::endl;
}

void print_pre_core_switch_message()
{
	std::cout<< "############ Switching off all cores (Expect Errors) ##################" << std::endl 
	<<std::endl;
}

void print_post_core_switch_message()
{
	std::cout << std::endl << "#######################################################################" << std::endl;
    	         std::cout << "########### Running Initialisation to find voltage level ##############" << std::endl;
}

void print_interrupt_message(int global_timer, bool high)
{
	if(high)
		std::cout << global_timer << " | High Voltage Interrupt Occured" << std::endl;
	else
		std::cout << global_timer << " | Low Voltage Interrupt Occured" << std::endl;
}

void print_frequency_message(int global_timer, bool high, int freq)
{
	if(VERBOSITY > 0)
	{
		if(high)
			std::cout << "   |-> Frequency Incremented to -> "<< freq << std::endl;
		else
			std::cout << "   |-> Frequency Decremented to -> "<< freq << std::endl;
	}
}

void print_to_log(std::ofstream &log_file, bool high, int small_cores, int big_cores, int freq)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	if(high)
		log_file << "0,1," << tv.tv_sec << "." << tv.tv_usec << "," << small_cores << "," << big_cores << "," << freq << std::endl;
}

void print_exit_message()
{
	std::cout << "Timeout Occured - Ending Session ..." << std::endl;
}

void print_hotplugging_message(bool adding, bool big)
{
	if(VERBOSITY > 0)
	{
		if(adding)
		{
			if(big)
				std::cout << "Adding a big core online "<< std::endl;
			else
				std::cout << "Adding a little core online "<< std::endl;
		}
		else
		{
			if(big)
				std::cout << "Removing a big core"<< std::endl;
			else
				std::cout << "Removing a small core"<< std::endl;
		}
	}
}

void print_seperator()
{
	std::cout<< "#######################################################################" << std::endl;
}

void print_max_limit(bool big_core)
{
	if(VERBOSITY > 1)
	{
		if(big_core)
			std::cout << "Maximum Cores reached"<< std::endl;
		else
			std::cout << "Maximum Small Cores reached"<< std::endl;
	}
}

void print_min_limit(bool big_core)
{
	if(VERBOSITY > 1)
	{
		if(big_core)
			std::cout << "Minimum Cores reached"<< std::endl;
		else
			std::cout << "Minimum Small Cores reached"<< std::endl;
	}
}

void print_success(int high, int low)
{
	print_seperator();
	std::cout << std::endl << "       Success! High Threshold | " << high << " | Low Threshold | " << low << std::endl;
	print_seperator();
	std::cout<< std::endl << "##################### Running Main Algorithm ##########################" << std::endl;
}
